#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include "err.h"

#define NR_PROC 5
#define NUMBER_L 6

int main (int argc, char* argv[])
{
  pid_t  pid, mypid;
  int    i, j, maxdesc, howmany;
  int    pipe_fd[2];
  fd_set desc_set;
  int    desc_arr[NR_PROC];
  char   number[NUMBER_L];

  printf("My process id = %d\n", getpid());

  /* clear the descriptors set */ 
  FD_ZERO(&desc_set);
  /* the max descriptor number in the set */
  maxdesc = 0;

  /* create NR_PROC child processes and remember read descriptors
     numbers in desc_set, desc_arr and the max one on maxdesc      */
  for (i = 0; i <= NR_PROC-1; i++) {
      
    /* create pipe for the next child process */
    if (pipe (pipe_fd) == -1) 
      syserr("Error in pipe\n");

      /* create child process */
      switch (pid = fork()) {
        case -1:
          syserr("Error in fork\n");
	  break;
	
        case 0: /* in child process */
          /* close pipe for reading */
	  close(pipe_fd[0]);
	  
	  /* print process info */
	  mypid = getpid();
	  printf("Child process: My process id = %d\n", mypid);
	  printf("Child process: Value returned by fork() = %d\n", pid);
	  
	  /* sleep for random no of seconds */
	  srand(mypid);
	  sleep(rand() % 5 + 1);
	  
	  /* write the pid to the pipe */ 
	  sprintf(number, "%d5", mypid);
	  if (write(pipe_fd[1], number, NUMBER_L - 1) < 0)
	    syserr("Error in write\n");
	    
	  /* end of process life */
	  exit(0);
	  break;
	
        default: /* parent process */
          /* close pipe for writing */
	  close(pipe_fd[1]);
	  
	  /* set desc_set, desc_arr and maxdesc */
	  FD_SET(pipe_fd[0], &desc_set);
	  desc_arr[i] = pipe_fd[0];
	  if (maxdesc < pipe_fd[0])
	    maxdesc = pipe_fd[0];
	  break;
	  
      } /* switch */ 
  } /* for */

  /* read info from all processes */
  j = 0;
  while (j < NR_PROC) {
    /* wait for reading on given descriptors set */ 
    howmany = select(maxdesc + 1, &desc_set, 0, 0, 0);
    if (howmany < 0)
      syserr("Error in select\n");
    /* there are howmany descriptors ready to read from */
    j += howmany;
      
    /* for each descriptor check if there was change in
       it's place in the set;
       if yes - read data from it and print it */
    for (i = 0; i < NR_PROC; i++) {
      /* did we read from this one already */
      if (desc_arr[i] > 0) /* if no */
        /* was it set by select function */
	if (FD_ISSET(desc_arr[i], &desc_set)) { /* if yes */
	  /* read form this descriptor */
	  howmany = read(desc_arr[i], number, NUMBER_L - 1);
	  if (howmany < 0)
	    syserr("Error in reading from pipe\n");
	    
	  /* print the data from subprocess */
	  number[howmany] = '\0';
	  printf("Value from the process %s\n", number);
	  
	  /* unset the number from descriptor */
	  FD_CLR(desc_arr[i], &desc_set);
	  
	  /* close the descriptor - we don't need it */
	  close(desc_arr[i]);
	  
	  /* set -1 in desc_arr */
	  desc_arr[i] = -1;
	}
    }
    
    /* clear the desc_set */
    FD_ZERO(&desc_set);
    /* and set only descriptors from which we did'n read */
    for (i = 0; i < NR_PROC; i++)
      if (desc_arr[i] > 0)
	FD_SET(desc_arr[i], &desc_set);
  }
  
  /* wait for all the processes */
  for (i = 0; i < NR_PROC; i++)
    if (wait(0) == -1) 
      syserr("Error in wait\n");
      
  /* end of the main process */
  exit(0);
}
